// Copyright 2020 The Brave Authors. All rights reserved.
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import Foundation
import Lottie
import BraveUI
import Shared

enum WelcomeCardAction {
    case closedButtonTapped
    case settingsButtonTapped
    case learnMoreButtonTapped
}

class BraveTodayWelcomeView: UIView, FeedCardContent {
    private let backgroundView = FeedCardBackgroundView()
    
    private let stackView = UIStackView().then {
        $0.axis = .vertical
        $0.alignment = .center
        $0.spacing = 16
    }
    
    let graphicAnimationView = AnimationView(name: "brave-today-welcome-graphic").then {
        $0.contentMode = .scaleAspectFit
        $0.loopMode = .loop
    }
    
    var introCardActionHandler: ((WelcomeCardAction) -> Void)?
    
    private let closeButton = UIButton(type: .system).then {
        $0.setImage(#imageLiteral(resourceName: "card_close").withRenderingMode(.alwaysOriginal), for: .normal)
        $0.accessibilityLabel = Strings.close
    }
    
    private let sourcesAndSettingsButton = ActionButton(type: .system).then {
        $0.layer.borderWidth = 0
        $0.titleLabel?.font = .systemFont(ofSize: 16.0, weight: .semibold)
        $0.setTitle(Strings.BraveToday.sourcesAndSettings, for: .normal)
        $0.contentEdgeInsets = UIEdgeInsets(top: 10, left: 20, bottom: 10, right: 20)
        $0.backgroundColor = UIColor.white.withAlphaComponent(0.2)
    }
    
    private let learnMoreButton = UIButton(type: .system).then {
        $0.setTitle(Strings.BraveToday.learnMoreTitle, for: .normal)
        $0.titleLabel?.font = .systemFont(ofSize: 15.0, weight: .semibold)
        $0.titleLabel?.appearanceTextColor = .white
        $0.setTitleColor(.white, for: .normal)
    }
    
    required init() {
        super.init(frame: .zero)
        
        addSubview(backgroundView)
        addSubview(stackView)
        addSubview(closeButton)
        
        backgroundView.snp.makeConstraints {
            $0.edges.equalToSuperview()
        }
        stackView.snp.makeConstraints {
            $0.edges.equalToSuperview().inset(24)
        }
        stackView.addStackViewItems(
            .view(graphicAnimationView),
            .customSpace(30),
            .view(UILabel().then {
                $0.text = Strings.BraveToday.introCardTitle
                $0.textAlignment = .center
                $0.appearanceTextColor = .white
                $0.font = .systemFont(ofSize: 18, weight: .semibold)
                $0.numberOfLines = 0
            }),
            .view(UILabel().then {
                $0.text = Strings.BraveToday.introCardBody
                $0.textAlignment = .center
                $0.appearanceTextColor = .white
                $0.font = .systemFont(ofSize: 14)
                $0.numberOfLines = 0
            }),
            .customSpace(24),
            .view(sourcesAndSettingsButton),
            .view(learnMoreButton)
        )
        
        closeButton.snp.makeConstraints {
            $0.top.right.equalToSuperview().inset(8)
        }
        
        closeButton.addTarget(self, action: #selector(tappedCloseButton), for: .touchUpInside)
        learnMoreButton.addTarget(self, action: #selector(tappedLearnMoreButton), for: .touchUpInside)
        sourcesAndSettingsButton.addTarget(self, action: #selector(tappedSettingsButton), for: .touchUpInside)
    }
    
    // MARK: - Actions
    
    @objc private func tappedCloseButton() {
        introCardActionHandler?(.closedButtonTapped)
    }
    
    @objc private func tappedLearnMoreButton() {
        introCardActionHandler?(.learnMoreButtonTapped)
    }
    
    @objc private func tappedSettingsButton() {
        introCardActionHandler?(.settingsButtonTapped)
    }
    
    @available(*, unavailable)
    required init(coder: NSCoder) {
        fatalError()
    }
    
    // MARK: - FeedCardContent
    
    var actionHandler: ((Int, FeedItemAction) -> Void)? {
        didSet {
            assertionFailure("Unused for welcome card")
        }
    }
    var contextMenu: FeedItemMenu? {
        didSet {
            assertionFailure("Unused for welcome card")
        }
    }
}
